from dataclasses import dataclass
from typing import Any

from django.http import HttpRequest

from routerific import router


@dataclass
class MethodGuard:
    method: str


def from_request(
    guard: MethodGuard,
    request: HttpRequest,
    context: router.RouteContext,
) -> Any:
    if request.method.lower() == guard.method.lower():
        return

    raise router.MatchFailure("Method not allowed")



from dataclasses import dataclass
from typing import Any

from django.http import HttpRequest

from routerific import router

from . import ParameterGuard


@dataclass
class PathGuard(ParameterGuard):
    def __post_init__(self):
        super().__post_init__()
        self.parser = router.build_parser(self.cls)


def from_request(
    guard: PathGuard,
    request: HttpRequest,
    context: router.RouteContext,
) -> Any:
    path_parameters = context.match.groupdict()

    # try to parse the path parameters into the view's parameter types
    try:
        value = path_parameters[guard.name]
        result = guard.parser(value)
        if result is None and guard.parameter.default is not None:
            return guard.parameter.default
        return result
    except Exception as e:
        raise router.MatchFailure("Failed to parse path parameter") from e



import datetime
import json
import uuid
from dataclasses import dataclass
from typing import Annotated, Type

import msgspec
import pytest
from django.http import HttpRequest

from routerific import Router, route
from routerific.expr import X
from routerific.guards import Header
from routerific.guards.parameter import ParameterGuard
from routerific.router import RouteConfigurationException, RouteContext


def test_match_path(rf):
    @route("get", "/blog/")
    def blog_list(): ...

    @route("get", r"/blog/(?P<id>\d+)")
    def blog_detail(id: int): ...

    request = rf.get("/blog/123")
    router = Router(
        views=[
            blog_list,
            blog_detail,
        ]
    )

    match = router.match(request)

    assert match is not None
    assert match.view == blog_detail


def test_match_path_django_syntax(rf):
    @route("get", "/blog/<int:id>")
    def blog_detail(id: int): ...

    request = rf.get("/blog/123")
    router = Router(views=[blog_detail])

    match = router.match(request)

    assert match is not None
    assert match.view == blog_detail


def test_match_method(rf):
    @route("get", "/blog/")
    def blog_list(): ...

    request = rf.post("/blog/")
    router = Router(views=[blog_list])

    match = router.match(request)

    assert match is None


def test_parameter_name_mismatch():

    with pytest.raises(RouteConfigurationException) as e:

        @route("get", r"/blog/(?P<post_id>\d+)")
        def view_func(id): ...

        Router(views=[view_func])

    assert "Path parameter named 'post_id' not found in view function" in str(e)


def test_parameter_type_parse_regular(rf):
    @route("get", r"/blog/(?P<id>\d+)")
    def blog_detail(id: int): ...

    router = Router(views=[blog_detail])

    request = rf.get("/blog/123")
    match = router.match(request)

    assert match.args == {"id": 123}


def test_parameter_type_parse_failure(rf):
    @route("get", r"/blog/(?P<id>.*)")
    def blog_detail(id: int): ...

    router = Router(views=[blog_detail])

    request = rf.get("/blog/abc")
    match = router.match(request)

    assert match is None


def test_unsupported_parameter_type(rf):
    class UnsupportedType: ...

    with pytest.raises(RouteConfigurationException) as e:

        @route("get", r"/blog/(?P<id>\d+)")
        def blog_detail(id: UnsupportedType): ...

        Router(views=[blog_detail])

    assert "Unsupported type 'UnsupportedType'" in str(e)


def test_path_parameter_type_default_str(rf):
    @route("get", r"/blog/(?P<id>\d+)")
    def view_func(id): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/123")
    match = router.match(request)

    assert match.args == {"id": "123"}


@pytest.mark.parametrize(
    "path_param,path_cls,url_param,result",
    [
        ("<int:id>", int, "123", 123),
        ("<str:id>", str, "123", "123"),
        ("<path:id>", str, "abc/def", "abc/def"),
        (
            "<uuid:id>",
            uuid.UUID,
            "6ba7b810-9dad-11d1-80b4-00c04fd430c8",
            uuid.UUID("6ba7b810-9dad-11d1-80b4-00c04fd430c8"),
        ),
        (
            "<slug:id>",
            str,
            "abc-def",
            "abc-def",
        ),
    ],
)
def test_path_parameter_django_syntax(rf, path_param, path_cls, url_param, result):
    @route("get", f"/blog/{path_param}")
    def view_func(id: path_cls): ...

    router = Router(views=[view_func])

    request = rf.get(f"/blog/{url_param}")
    match = router.match(request)

    assert match.args == {"id": result}


def test_overload_single_function(rf):
    @route("post", "/blog/")
    @route("get", "/blog/")
    def view_func(): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match.view is view_func

    request = rf.post("/blog/")
    match = router.match(request)
    assert match.view is view_func


def test_multiple_path_parameters(rf):
    @route("get", r"/shop/(?P<shop_id>\d+)/products/(?P<product_id>\d+)")
    def view_func(shop_id: int, product_id: int): ...

    router = Router(views=[view_func])

    request = rf.get("/shop/123/products/456")
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"shop_id": 123, "product_id": 456}


def test_body_parameter_regular(rf):

    class Input(msgspec.Struct):
        date: datetime.date

    @route("post", r"/blog/")
    def view_func(x: Input): ...

    router = Router(
        views=[view_func],
        integrations=["routerific.guards.integrations.msgspec.from_request"],
    )

    request = rf.post(
        "/blog/",
        data=json.dumps({"date": "2023-01-01"}),
        content_type="application/json",
    )
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"x": Input(date=datetime.date(2023, 1, 1))}


def test_body_parameter_parse_failure(rf):
    class Input(msgspec.Struct):
        a: int

    @route("post", r"/blog/")
    def view_func(x: Input): ...

    router = Router(
        views=[view_func],
        integrations=["msgspec"],
    )

    request = rf.post(
        "/blog/",
        data=json.dumps({"a": "not-an-int"}),
        content_type="application/json",
    )
    match = router.match(request)
    assert match is None


def test_body_parameter_parse_multiple(rf):
    class InputA(msgspec.Struct):
        a: int

    class InputB(msgspec.Struct):
        b: int

    @route("post", r"/blog/")
    def view_func(x: InputA, y: InputB): ...

    router = Router(views=[view_func], integrations=["msgspec"])

    request = rf.post(
        "/blog/",
        data=json.dumps({"a": 1, "b": 2}),
        content_type="application/json",
    )

    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"x": InputA(a=1), "y": InputB(b=2)}


def test_query_parameter_regular(rf):
    @route("get", r"/blog/")
    def view_func(a: int): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/?a=123")
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"a": 123}


def test_query_parameter_parse_failure(rf):
    @route("get", r"/blog/")
    def view_func(a: int): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/?a=abc")
    match = router.match(request)
    assert match is None


def test_query_parameter_not_found(rf):
    @route("get", r"/blog/")
    def view_func(a: int): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match is None


@pytest.mark.parametrize(
    "cls,constraint,param,found",
    [
        (int, X >= 0, 123, True),
        (int, X <= 100, 123, False),
        (str, X.length() > 4, "abc", False),
        (str, X.length() < 4, "abc", True),
    ],
)
def test_query_parameter_expr_constraint_int(rf, cls, constraint, param, found):
    @route("get", r"/blog/")
    def view_func(a: Annotated[cls, constraint]): ...

    router = Router(views=[view_func])

    request = rf.get(f"/blog/?a={param}")
    match = router.match(request)
    assert (getattr(match, "view", None) is view_func) == found


@pytest.mark.parametrize(
    "name,constraint,found",
    [
        ("a", X == 123, True),
        ("a", X == 456, False),
        ("b", X == 123, False),
    ],
)
def test_explicit_header_guard_annotation(rf, name, constraint, found):
    @route("get", r"/blog/")
    def view_func(a: Annotated[int, Header, constraint]): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/", headers={name: "123"})

    match = router.match(request)
    assert (match is not None) == found


@pytest.mark.parametrize(
    "value,found",
    [
        ("0", False),
        ("10", True),
    ],
)
def test_multiple_constraints(rf, value, found):
    @route("get", r"/blog/")
    def view_func(a: Annotated[int, X > 0, X < 100, Header]): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/", headers={"a": value})

    match = router.match(request)
    assert (match is not None) == found


def test_custom_matcher(rf):
    @dataclass
    class User:
        id: str

    def user_guard(guard: User, request: HttpRequest, context: RouteContext) -> User:
        return User(id=request.headers["user"])

    @route("get", r"/blog/")
    def view_func(user: User): ...

    router = Router(views=[view_func], integrations=[user_guard])

    request = rf.get("/blog/", headers={"user": "user-id-123"})

    match = router.match(request)
    assert match is not None
    assert match.view is view_func
    assert match.args == {"user": User(id="user-id-123")}


def test_multiple_integrations(rf):
    class UserHeader(msgspec.Struct):
        id: str

    class UserBody(msgspec.Struct):
        id: str

    def user_guard(
        guard: ParameterGuard[UserHeader], request: HttpRequest, context: RouteContext
    ) -> UserHeader:
        return UserHeader(id=request.headers["user"])

    @route("post", r"/blog/")
    def view_func(user_header: UserHeader, user_body: UserBody): ...

    router = Router(views=[view_func], integrations=[user_guard, "msgspec"])

    request = rf.post(
        "/blog/",
        headers={"user": "user-id-123"},
        data=json.dumps({"id": "user-id-456"}),
        content_type="application/json",
    )

    match = router.match(request)
    assert match is not None
    assert match.view is view_func
    assert match.args["user_header"] == UserHeader(id="user-id-123")
    assert match.args["user_body"] == UserBody(id="user-id-456")


def test_integration_variance(rf):
    class User(msgspec.Struct):
        id: str

    class Item(msgspec.Struct):
        id: str

    def from_request(
        guard: ParameterGuard[User],
        request: HttpRequest,
        context: RouteContext,
    ) -> User:
        return User(id=request.headers["user"])

    @route("post", r"/blog/")
    def view_func(user: User, item: Item): ...

    router = Router(views=[view_func], integrations=["msgspec", from_request])

    request = rf.post(
        "/blog/",
        headers={"user": "user-id-123"},
        data=json.dumps({"id": "item-id-456"}),
        content_type="application/json",
    )

    match = router.match(request)
    assert match is not None
    assert match.view is view_func
    assert match.args["user"] == User(id="user-id-123")
    assert match.args["item"] == Item(id="item-id-456")


def test_optional_query_parameter(rf):
    @route("get", r"/blog/")
    def view_func(a: int | None = None): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"a": None}


def test_optional_query_parameter_with_default(rf):
    @route("get", r"/blog/")
    def view_func(a: int | None = 123): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"a": 123}


def test_optional_header_parameter(rf):
    @route("get", r"/blog/")
    def view_func(a: Annotated[int | None, Header] = None): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match.view is view_func


def test_optional_header_parameter_with_default(rf):
    @route("get", r"/blog/")
    def view_func(a: Annotated[int | None, Header] = 123): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/")
    match = router.match(request)
    assert match.view is view_func
    assert match.args == {"a": 123}


def test_optional_path_parameter(rf):
    @route("get", r"/((?P<lang>\d+)/)?blog/<int:id>/")
    def view_func(id: int, lang: str | None = None): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/123/")
    match = router.match(request)
    assert match.view is view_func
    assert match.args["id"] == 123
    assert match.args["lang"] is None


def test_optional_path_parameter_with_default(rf):
    @route("get", r"/((?P<lang>\d+)/)?blog/<int:id>/")
    def view_func(id: int, lang: str | None = "en"): ...

    router = Router(views=[view_func])

    request = rf.get("/blog/123/")
    match = router.match(request)
    assert match.view is view_func
    assert match.args["id"] == 123
    assert match.args["lang"] == "en"



import pytest

from routerific import Router, route
from routerific.router import RouteConfigurationException


def test_empty_router(rf):
    router = Router()

    response = router.dispatch(rf.get("/"))
    assert response.status_code == 404


def test_dispatch_regular(rf):
    result = {}

    @route("get", "/blog/")
    def blog_list(): ...

    @route("get", r"/blog/(?P<id>\d+)")
    def blog_detail(id: int):
        return result

    router = Router(views=[blog_list, blog_detail])
    response = router.dispatch(rf.get("/blog/123"))

    assert response is result


def test_dispatch_not_found(rf):
    @route("get", "/blog/")
    def blog_list(): ...

    router = Router(views=[blog_list])
    response = router.dispatch(rf.get("/blub/123"))
    assert response.status_code == 404


def test_route_not_annotated(rf):
    def blog_list(): ...

    with pytest.raises(RouteConfigurationException) as e:
        Router(views=[blog_list])

    assert "View function must be annotated with @route" in str(e)



from typing import Any

import msgspec
from django.http import HttpRequest

from routerific import router
from routerific.guards.parameter import ParameterGuard


def from_request(
    guard: ParameterGuard[msgspec.Struct],
    request: HttpRequest,
    context: router.RouteContext,
) -> Any:
    try:
        result = msgspec.json.decode(request.body, type=guard.parameter.annotation)
        return result
    except (msgspec.ValidationError, msgspec.DecodeError) as e:
        raise router.MatchFailure(f"Invalid body: {e}") from e



from typing import Any

from django.http import HttpRequest

from routerific import router

from . import ParameterGuard


class HeaderGuard(ParameterGuard):
    def __post_init__(self, *args, **kwargs):
        super().__post_init__(*args, **kwargs)
        self.parser = router.build_parser(self.cls)


def from_request(
    guard: HeaderGuard,
    request: HttpRequest,
    context: router.RouteContext,
) -> Any:
    value = request.headers.get(guard.name.upper())

    try:
        result = guard.parser(value)
        if result is None and guard.parameter.default is not None:
            return guard.parameter.default
        return result
    except Exception as e:
        raise router.MatchFailure(f"Failed to parse header {guard.name!r}: {e}") from e



"""Terrific routing for Django."""

__version__ = "0.1.0"

from typing import TypeAlias

from routerific.guards import (
    HeaderGuard,
    MethodGuard,
    PathGuard,
    QueryGuard,
)

from .router import Router, route



import inspect
from dataclasses import dataclass
from typing import Generic, Type, TypeVar

T = TypeVar("T")


@dataclass
class ParameterGuard(Generic[T]):
    parameter: inspect.Parameter
    cls: Type[T]

    def __post_init__(self):
        self.name = self.parameter.name



from typing import Any


class Expr:
    def evaluate(self, variables: dict[str, Any]) -> Any:
        raise NotImplementedError

    def compile(self):
        return eval(f"lambda x: {self}")

    def __str__(self) -> str:
        raise NotImplementedError

    def __add__(self, other):
        return Add(self, other)

    def __sub__(self, other):
        return Sub(self, other)

    def __mul__(self, other):
        return Mul(self, other)

    def __truediv__(self, other):
        return Div(self, other)

    def __neg__(self):
        return Neg(self)

    def length(self):
        return Length(self)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Ne(self, other)

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)


class Add(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Add({self.left!r}, {self.right!r})"

    def __str__(self):
        return f"({self.left} + {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) + evaluate(self.right, variables)


class Sub(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Sub({self.left!r}, {self.right!r})"

    def __str__(self):
        return f"({self.left} - {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) - evaluate(self.right, variables)


class Mul(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Mul({self.left!r}, {self.right!r})"

    def __str__(self):
        return f"({self.left} * {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) * evaluate(self.right, variables)


class Div(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Div({self.left!r}, {self.right!r})"

    def __str__(self):
        return f"({self.left}/{self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) / evaluate(self.right, variables)


class Neg(Expr):
    def __init__(self, expr):
        self.expr = expr

    def __repr__(self):
        return f"Neg({self.expr!r})"

    def __str__(self) -> str:
        return f"-{self.expr}"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return -evaluate(self.expr, variables)


class Length(Expr):
    def __init__(self, expr):
        self.expr = expr

    def __repr__(self):
        return f"Len({self.expr!r})"

    def __str__(self) -> str:
        return f"len({self.expr})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return len(evaluate(self.expr, variables))


class Gt(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Gt({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} > {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) > evaluate(self.right, variables)


class Ge(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Gte({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} >= {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) >= evaluate(self.right, variables)


class Lt(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Lt({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} < {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) < evaluate(self.right, variables)


class Le(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Lte({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} <= {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) <= evaluate(self.right, variables)


class Eq(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Eq({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} == {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) == evaluate(self.right, variables)


class Ne(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Ne({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} != {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) != evaluate(self.right, variables)


class And(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"And({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} and {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) and evaluate(self.right, variables)


class Or(Expr):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Or({self.left!r}, {self.right!r})"

    def __str__(self) -> str:
        return f"({self.left} or {self.right})"

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return evaluate(self.left, variables) or evaluate(self.right, variables)


class Var(Expr):
    def __init__(self, name):
        if not str.isidentifier(name):
            raise ValueError(f"Invalid variable name: {name}")
        self.name = name

    def __repr__(self):
        return f"Var({self.name!r})"

    def __str__(self) -> str:
        return self.name

    def evaluate(self, variables: dict[str, Any]) -> Any:
        return variables[self.name]


def evaluate(expr: Any, variables: dict[str, Any]) -> Any:
    if isinstance(expr, Expr):
        return expr.evaluate(variables)
    elif isinstance(expr, (str, int, float, bool, type(None))):
        return expr

    raise ValueError(f"Unsupported expr {type(expr)}")


X = Var("x")



from typing import Annotated

from django.contrib import admin
from django.http import HttpResponse
from django.urls import path, re_path

from routerific import Router, route
from routerific.expr import X
from routerific.router import include


@route("get", "/blog/")
def list_view():
    return HttpResponse("Hello World")


@route("get", "/blog/<str:id>")
def detail_view(
    id: Annotated[str, X.length() > 4],
):
    return HttpResponse(f"Hello {id}")


views = [
    list_view,
    detail_view,
]


router = Router(
    [
        include("example.urls.views"),
    ]
)


urlpatterns = [
    path("admin/", admin.site.urls),
    re_path("^blog", router.dispatch, name="blog"),
]



import inspect
from typing import Annotated, TypeAlias

import pytest

from routerific.expr import Var, X, evaluate


def test_simple_str():
    expr = X + X

    assert str(expr) == "(x + x)"


def test_length():
    expr = X.length() > 10

    assert str(expr) == "(len(x) > 10)"


def test_complex():
    """Some non-sensical expressions to test every operator."""
    Y = Var("y")
    expr_1 = (-X * X) / (X > 0)
    expr_2 = (Y.length() < 1) + (X == X) - (X != Y)
    expr = expr_1 & (expr_2 <= 1) | (expr_2 > 1) | (expr_2 >= 0)

    assert str(expr)
    assert repr(expr)

    assert evaluate(expr, {"x": 1, "y": "abc"}) is not None


def test_unsupported_type():
    with pytest.raises(ValueError):
        evaluate({}, {})


def test_simple_eval():
    expr = X + X
    variables = {"x": 1}

    assert evaluate(expr, variables) == 2


def test_simple_cond():
    expr = X + 1 >= 2
    variables = {"x": 1}

    assert evaluate(expr, variables) is True


def test_annotated():
    def f(x: Annotated[int, (X >= 0) & (X < 10)]): ...

    signature = inspect.signature(f)
    parameters = signature.parameters

    assert "x" in parameters

    meta = parameters["x"].annotation.__metadata__
    assert len(meta) == 1

    expr = meta[0]

    assert str(expr) == "((x >= 0) and (x < 10))"

    # Now for some crazy stuff
    assert eval(f"lambda x: {expr}")(1) is True
    assert eval(f"lambda x: {expr}")(11) is False


def test_positive_int_annotation():
    PositiveInt: TypeAlias = Annotated[int, X >= 0]

    def f(x: PositiveInt): ...

    signature = inspect.signature(f)
    parameters = signature.parameters

    assert "x" in parameters

    meta = parameters["x"].annotation.__metadata__
    assert len(meta) == 1

    expr = meta[0]

    assert str(expr) == "(x >= 0)"


def test_var_name_invalid():
    with pytest.raises(ValueError):
        Var("from os import *; os.system('rm -rf /')")



import importlib
import inspect
import re
import typing
from dataclasses import dataclass
from functools import reduce
from types import NoneType
from typing import Any, Callable, Iterable
from uuid import UUID

from django.http import HttpResponseNotFound
from more_itertools import collapse

import routerific.guards as guards
from routerific.guards.parameter import ParameterGuard

from .expr import Expr

GUARDS_ATTR = "_guards"


class MatchFailure(Exception):
    pass


class RouteConfigurationException(Exception):
    pass


def parse_str(value):
    if value is None:
        raise ValueError
    return str(value)


DESERIALIZERS = {
    int: int,
    str: parse_str,
    UUID: UUID,
    NoneType: lambda _: None,
}

DJANGO_PATTERNS = {
    "int": r"\d+",
    "str": r"[^/]+",
    "slug": r"[a-zA-Z0-9-_]+",
    "uuid": r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
    "path": r".+",
}


def path_to_pattern(path: str) -> re.Pattern:
    def replacer(match):
        name = match.group("name")
        type_ = match.group("type")
        pattern = DJANGO_PATTERNS[type_]
        return f"(?P<{name}>{pattern})"

    # This regex tries to match django-style path parameters, but
    # not regex-style named capture groups. This seems somewhat brittle,
    # but we'll go with it for now.
    param_regex = re.compile(r"(?<!\?P)<((?P<type>[^:]+):)?(?P<name>[^>]+)>")
    regexed_pattern = param_regex.sub(replacer, path)
    return re.compile(regexed_pattern)


def path_parameter_names(path: str) -> list[str]:
    pattern = path_to_pattern(path)
    return list(pattern.groupindex.keys())


def build_parser(cls: type) -> Callable[[str], Any]:
    predicate = None

    if typing.get_origin(cls) is typing.Annotated:
        args = typing.get_args(cls)
        cls = args[0]

        exprs = [expr for expr in args[1:] if isinstance(expr, Expr)]
        if exprs:
            expr = reduce(lambda a, b: a & b, exprs)
            predicate = expr.compile()
    else:
        cls = cls

    try:
        if args := typing.get_args(cls):
            deserializers = [DESERIALIZERS[arg] for arg in args]
        else:
            deserializers = [DESERIALIZERS[cls]]
    except KeyError:
        raise RouteConfigurationException(f"Unsupported type {cls.__name__!r}")

    def parse(value):
        for deserializer in deserializers:
            try:
                deserialized_value = deserializer(value)
                if predicate is None:
                    return deserialized_value

                if not predicate(deserialized_value):
                    continue

                return deserialized_value
            except Exception:
                continue

        raise ValueError

    return parse


@dataclass
class RouteContext:
    match: re.Match


@dataclass
class RouterMatch:
    view: Callable[..., Any]
    args: dict[str, Any]


@dataclass
class ViewGuard:
    view: Callable[..., Any]
    guards: list[guards.ParameterGuard]
    path_pattern: re.Pattern


@dataclass
class Router:
    def __init__(
        self,
        views: list[Callable[..., Any]] | None = None,
        integrations: list[str | Callable[..., Any]] | None = None,
    ):
        self.guard_registry = {}

        # register standard types
        from routerific.guards import header, method, path, query

        self.register_guard(header.from_request)
        self.register_guard(method.from_request)
        self.register_guard(path.from_request)
        self.register_guard(query.from_request)

        if integrations is not None:
            for integration in integrations:
                self.register_guard(integration)

        self.view_guards = list(collapse(map(self.view_guard, views or [])))

    def register_guard(self, fn: str | Callable[..., Any]) -> None:
        if isinstance(fn, str):
            fn_split = fn.rsplit(".", 1)
            if len(fn_split) == 1:
                module_name = f"routerific.guards.integrations.{fn}"
                attr_name = "from_request"
            else:
                module_name, attr_name = fn_split

            module = importlib.import_module(module_name)
            fn = getattr(module, attr_name)

        signature = inspect.signature(fn)
        cls = next(iter(signature.parameters.values())).annotation

        if typing.get_origin(cls) == ParameterGuard and (args := typing.get_args(cls)):
            cls = args[0]

        self.guard_registry[cls] = fn

    def view_guard(self, view: Callable[..., Any]) -> Iterable[ViewGuard]:
        routerific = getattr(view, "__routerific__", None)
        if routerific is None:
            raise RouteConfigurationException(
                "View function must be annotated with @route"
            )

        for method, path in routerific:
            path_pattern = path_to_pattern(path)
            view_guard = ViewGuard(
                guards=self.view_guards(view, method, path),
                path_pattern=path_pattern,
                view=view,
            )

            # check if all path parameters are actually instantiated as path guards
            for name in path_pattern.groupindex.keys():
                for guard in view_guard.guards:
                    if isinstance(guard, guards.PathGuard) and guard.name == name:
                        break
                else:
                    raise RouteConfigurationException(
                        f"Path parameter named {name!r} not found in view function"
                    )

            yield view_guard

    def _from_request(self, guard, request, context):
        if not isinstance(guard, guards.ParameterGuard):
            fn = self.guard_registry[guard.__class__]
            return fn(guard, request, context)

        cls = guard.parameter.annotation
        is_guard = isinstance(guard, guards.ParameterGuard)
        is_type = isinstance(cls, type)

        if is_guard and is_type and (fn := self.guard_registry.get(guard.cls)):
            return fn(guard, request, context)

        fn = self.guard_registry[guard.__class__]
        return fn(guard, request, context)

    def match(self, request) -> RouterMatch | None:
        candidates = []

        for view_guard in self.view_guards:
            match = view_guard.path_pattern.fullmatch(request.path)
            if match is None:
                continue

            context = RouteContext(match=match)
            try:
                args = {}
                for guard in view_guard.guards:
                    value = self._from_request(guard, request, context)
                    if isinstance(guard, guards.ParameterGuard):
                        args[guard.name] = value

                candidates.append(RouterMatch(view_guard.view, args))
            except MatchFailure:
                continue

        return next(iter(candidates), None)

    def dispatch(self, request):
        match = self.match(request)
        if match is None:
            return HttpResponseNotFound()

        return match.view(**match.args)

    @staticmethod
    def include(path: str):
        module, member = path.rsplit(".", 1)
        module = importlib.import_module(module)
        views = getattr(module, member)
        assert isinstance(views, list)
        return views

    @staticmethod
    def cls_or_str(cls: type) -> type:
        if cls is inspect.Parameter.empty:
            return str
        return cls

    def implicitly_located_guard(
        self, path: str, parameter: inspect.Parameter
    ) -> guards.ParameterGuard:
        cls = Router.cls_or_str(parameter.annotation)
        if parameter.name in path_parameter_names(path):
            return guards.PathGuard(parameter=parameter, cls=cls)

        if isinstance(cls, type):
            if self.guard_registry.get(cls) is not None:
                return guards.ParameterGuard(parameter=parameter, cls=cls)

            if self.guard_registry.get(cls.__bases__[0]) is not None:
                cls = cls.__bases__[0]
                return guards.ParameterGuard(parameter=parameter, cls=cls)

        return guards.QueryGuard(parameter=parameter, cls=cls)

    def explicitly_located_guard(
        self, parameter: inspect.Parameter
    ) -> guards.ParameterGuard:
        args = typing.get_args(parameter.annotation)
        cls = parameter.annotation
        for arg in args:
            if self.is_explicit_guard(arg):
                return arg(parameter=parameter, cls=cls)

        return None

    @staticmethod
    def is_explicit_guard(cls) -> bool:
        for guard in (
            guards.HeaderGuard,
            guards.QueryGuard,
            guards.PathGuard,
        ):
            if cls is guard:
                return True

        return False

    def parameter_guard(
        self, *, parameter: inspect.Parameter, path: str
    ) -> guards.ParameterGuard:
        if guard := self.explicitly_located_guard(parameter=parameter):
            return guard

        return self.implicitly_located_guard(parameter=parameter, path=path)

    def view_guards(
        self, view: Callable[..., Any], method: str, path: str
    ) -> list[guards.ParameterGuard]:
        signature = inspect.signature(view)

        return [
            guards.MethodGuard(method=method),
            *(
                self.parameter_guard(path=path, parameter=parameter)
                for parameter in signature.parameters.values()
            ),
        ]


def route(method, path: str):
    def decorator(view):
        if not hasattr(view, "__routerific__"):
            setattr(view, "__routerific__", [(method, path)])
        else:
            getattr(view, "__routerific__").append((method, path))

        return view

    return decorator



from typing import TypeAlias

from routerific.guards.parameter import ParameterGuard

from .header import HeaderGuard
from .method import MethodGuard
from .path import PathGuard
from .query import QueryGuard

Query: TypeAlias = QueryGuard
Method: TypeAlias = MethodGuard
Path: TypeAlias = PathGuard
Header: TypeAlias = HeaderGuard
